"""Base class for Avro Preprocessor modules"""

from abc import abstractmethod, ABC
from collections import OrderedDict
from dataclasses import dataclass
from typing import Callable, Generator, List, Optional, Tuple, Union

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


@dataclass(frozen=True)  # pylint: disable=R0903
class Ancestor:
    """A data class for JSON ancestors"""
    node: Avro.Node
    key: Union[str, int]


class PreprocessorModule(ABC):
    """Base class for Avro Preprocessor modules"""

    def __init__(self, schemas: SchemasContainer) -> None:
        self.schemas: SchemasContainer = schemas
        self.current_schema_name = ""

        self.ancestors: List[Ancestor] = []

    @abstractmethod
    def process(self) -> None:
        """Entry method to process data with a module."""

    def processed_schemas_and_keys_iter(self) \
            -> Generator[Tuple[str, OrderedDict], None, None]:
        """
        A generator to iterate over processed schemas and autogenerated keys.
        I also sets the current schema name so it can be retrieved by AvroPreprocessor
        in case an exception happens.
        """
        all_schemas = list(self.schemas.processed.items()) \
                      + list(self.schemas.autogenerated_keys.items())

        for schema_name, schema in all_schemas:
            self.current_schema_name = schema_name
            yield schema_name, schema

    def processed_schemas_iter(self) -> \
            Generator[Tuple[str, OrderedDict], None, None]:
        """
        A generator to iterate over processed schemas.
        I also sets the current schema name so it can be retrieved by AvroPreprocessor
        in case an exception happens.
        """
        for schema_name, schema in self.schemas.processed.items():
            self.current_schema_name = schema_name
            yield schema_name, schema

    def original_schemas_iter(self) -> Generator[Tuple[str, OrderedDict], None, None]:
        """
        A generator to iterate over original schemas.
        I also sets the current schema name so it can be retrieved by AvroPreprocessor
        in case an exception happens.
        """
        for schema_name, schema in self.schemas.original.items():
            self.current_schema_name = schema_name
            yield schema_name, schema

    def traverse_schema(
            self,
            func: Callable[[Avro.Node], Optional[bool]],
            node: Avro.Node,
    ) -> None:
        """
        Traverse a schema and apply 'func' to chosen elements.
        :param func: The function to apply, can return bool or None
        :param node: The [Avro.Node] to traverse
        """

        self.ancestors.clear()
        self.do_traverse_schema(func, node)

    def do_traverse_schema(
            self,
            func: Callable[[Avro.Node], Optional[bool]],
            node: Avro.Node,
    ) -> Optional[bool]:
        """
        Helper function to recursively traverse a schema and apply 'func' to chosen elements.
        Returns True if it changed the data (so it can be applied again).
        :param func: The function to apply, can return bool or None
        :param node: The [Avro.Node] to traverse
        :return: bool
        """

        res = func(node)

        if isinstance(node, (list, OrderedDict)):

            # modifying node[key] while iterating. Here do not use e.g.
            # for key, value in node.items():
            for key in range(len(node)) if isinstance(node, list) else node.keys():

                self.ancestors.append(Ancestor(node, key))

                while self.do_traverse_schema(func, node[key]):
                    pass

                self.ancestors.pop()

        return res



"""
A module to register Avro schemas to Confluent Schema Registry.
"""

import json
from collections import OrderedDict

import requests
from requests import Response

from avro_preprocessor.avro_naming import AvroNaming
from avro_preprocessor.colored_json import ColoredJson
from avro_preprocessor.preprocessor_module import PreprocessorModule

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class SchemaRegistrar(PreprocessorModule):
    """
    Register schemas to Confluent Schema Registry.
    """
    schema_registry_url: str = ''
    url: str = ''
    payload: str = ''

    def process(self) -> None:
        """Process all schemas."""

        # If a key_schema is defined, it means we have a default common key.
        default_common_key = self.schemas.paths.key_schema

        # In that case, we process it with the schema registry.
        if default_common_key and \
                self.schemas.paths.key_subject_name_strategy == "RecordNameStrategy":
            common_key_schema = self.schemas.processed[self.schemas.paths.key_schema]
            self.perform_request_to_schema_registry("", default_common_key, common_key_schema)

        current_topic = ""

        for name, schema in self.processed_schemas_iter():

            if not self.schemas.paths.is_rpc_resource(name):
                # RPC protocols cannot be registered to the schema registry

                topic = AvroNaming.get_topic(self.schemas.paths.base_namespace, name)
                value_subject_name = AvroNaming.get_subject_name_for_value(
                    self.schemas.paths.base_namespace, name)
                self.perform_request_to_schema_registry(topic, value_subject_name, schema)

                if self.schemas.paths.is_event_resource(name) and topic != current_topic:
                    key_subject_name = AvroNaming.get_subject_name_for_key(
                        self.schemas.paths.base_namespace,
                        AvroNaming.get_key_fully_qualified_name(name))

                    if default_common_key and self.schemas.paths.key_subject_name_strategy \
                            == "TopicRecordNameStrategy":
                        key_schema = self.schemas.processed[self.schemas.paths.key_schema]
                        self.perform_request_to_schema_registry(topic, key_subject_name, key_schema)
                    else:
                        # if we were requested to generate keys (KeysGenerator module active),
                        # and the key_subject_name_strategy is not TopicRecordNameStrategy,
                        # we also register them to the schema registry
                        key_name = AvroNaming.get_key_fully_qualified_name(name)
                        if key_name in self.schemas.autogenerated_keys:
                            key_schema = self.schemas.autogenerated_keys[key_name]
                            self.perform_request_to_schema_registry(
                                topic, key_subject_name, key_schema)

                current_topic = topic

    def perform_request_to_schema_registry(
            self, topic: str, subject_name: str, schema: OrderedDict) -> None:
        """
        Register a schema in the schema registry.
        :param topic: The Kafka topic
        :param subject_name: The subject name for this schema
        :param schema: The schema
        """
        schema_text = json.dumps(schema, indent=ColoredJson.json_indent)

        self.payload = "{ \"schema\": \"" \
                  + schema_text.replace("\"", "\\\"").replace("\t", "").replace("\n", "") \
                  + "\" }"
        # print("Payload: " + self.payload)

        self.url = self.get_request_url(subject_name)

        headers = {"Content-Type": "application/vnd.schemaregistry.v1+json"}

        if self.schemas.verbose:
            print("Schema Registry URL: " + self.schema_registry_url)
            print("Topic: " + topic)
            print("Subject name: " + subject_name)
            print("URL: " + self.url)
            print("Headers :", headers)

        req = requests.post(self.url, headers=headers, data=self.payload)
        self.process_response(req)

    def get_request_url(self, subject: str) -> str:
        """
        Returns the URL for this request
        :param subject: The Avro subject name
        :return: The full URL
        """
        return self.schema_registry_url + "/subjects/" + subject + "/versions"

    def process_response(self, response: Response) -> None:
        """
        Process the response.
        :param response: The response
        """
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            if self.schemas.verbose:
                print('Response: Success!', json.loads(response.content), '\n')
        else:
            print('\n\nFailure, request payload:\n', self.payload)
            print('Response: ', json.loads(response.content), '\n')
            response.raise_for_status()



from typing import Any

from ruamel.yaml import YAML, StringIO  # type: ignore


class ExtendedYAML(YAML):  # type: ignore
    def dump_str(self, data: Any, **kw: Any) -> str:
        stream = StringIO()
        YAML.dump(self, data, stream, **kw)
        string: str = stream.getvalue()
        stream.close()
        return string



"""
A module to pretty print colored JSON
"""
import json
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, cast, Optional, Union

from pygments import highlight
from pygments.formatters.terminal256 import TerminalTrueColorFormatter
from pygments.lexers.data import JsonLexer, YamlLexer

from avro_preprocessor.extended_yaml import ExtendedYAML

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


@dataclass
class ColoredJson:
    """
    Pretty prints JSON on the terminal
    """
    json_lexer = JsonLexer()
    yaml_lexer = YamlLexer()
    formatter = TerminalTrueColorFormatter(style='monokai')

    json_indent = 4
    yaml_indent = {"mapping": 2, "sequence": 4, "offset": 2}

    @classmethod
    def highlight_json(cls, data: Union[Dict, OrderedDict],
                       indent: Optional[int] = json_indent) -> str:
        """
        Get a colored, highlighted version of a JSON string.
        :param data: The json to highlight
        :param indent: How much to indent fields in the JSON
        """
        json_text = json.dumps(data, indent=indent)
        return cast(str, highlight(json_text, cls.json_lexer, cls.formatter))

    @classmethod
    def highlight_yaml(cls, data: Union[Dict, OrderedDict],
                       indent: Optional[dict] = None) -> str:
        """
        Get a colored, highlighted version of a YAML string.
        :param data: The yaml to highlight
        :param indent: Yaml fields indentation
        """
        if indent is None:
            indent = cls.yaml_indent
        yaml = ExtendedYAML()
        yaml.default_flow_style = False
        yaml.indent(**indent)
        yaml.Representer.add_representer(OrderedDict, yaml.Representer.represent_dict)
        yaml_text = yaml.dump(data)
        return cast(str, highlight(yaml_text, cls.yaml_lexer, cls.formatter))



"""
A module to create a map fully_qualified_class_name -> topic name.
"""
import json
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, List, Any

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.avro_naming import AvroNaming
from avro_preprocessor.colored_json import ColoredJson
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


@dataclass
class SchemaMappingFields:  # pylint: disable=R0902
    """
    Fields for the schema mapping file
    """
    topic = 'topic'

    value_fqcn = 'value-fqcn'
    value_subject = 'value-subject'

    default_key_fqcn = 'default-key-fqcn'
    default_key_subject = 'default-key-subject'

    partition_key_fqcn = 'partition-key-fqcn'
    partition_key_subject = 'partition-key-subject'

    partition_fields = 'partition-fields'
    user_id_fields = 'user-id-fields'


SMF = SchemaMappingFields


class SchemaMappingGenerator(PreprocessorModule):
    """
    Generates and saves the schema mapping.
    """

    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)
        self.current_schema_name: str = ""
        self.schema_mapping: Dict[Any, Any] = {}

    def process(self) -> None:
        """Process all schemas."""

        for name, schema in self.processed_schemas_iter():
            if self.schemas.paths.is_event_resource(name) and all(
                    not name.startswith(exclude)
                    for exclude in self.schemas.paths.schema_mapping_exclude
            ):
                self.schema_mapping[name] = OrderedDict((

                    (SMF.topic, AvroNaming.get_topic(self.schemas.paths.base_namespace, name)),

                    (SMF.value_fqcn, name),
                    (SMF.value_subject, AvroNaming.get_subject_name_for_value(
                        self.schemas.paths.base_namespace, name)),

                    # Default common key uses RecordNameStrategy in the schema registry
                    (SMF.default_key_fqcn, self.schemas.paths.key_schema),
                    (SMF.default_key_subject, self.schemas.paths.key_schema),

                    (SMF.partition_key_fqcn, AvroNaming.get_key_fully_qualified_name(name)),
                    (SMF.partition_key_subject, AvroNaming.get_subject_name_for_key(
                        self.schemas.paths.base_namespace,
                        AvroNaming.get_key_fully_qualified_name(name))),

                    (SMF.partition_fields, []),
                    (SMF.user_id_fields, [])

                ))

                self.current_schema_name = name

                self.traverse_schema(self.find_partition_keys, schema)
                self.schema_mapping[name][SMF.partition_fields] = \
                    sorted(self.schema_mapping[name][SMF.partition_fields])

                self.traverse_schema(self.find_logical_types, schema)
                self.schema_mapping[name][SMF.user_id_fields] = \
                    sorted(self.schema_mapping[name][SMF.user_id_fields])

        sorted_schema_mapping = OrderedDict(sorted(self.schema_mapping.items()))
        sorted_schema_mapping_text = \
            json.dumps(sorted_schema_mapping, indent=ColoredJson.json_indent)
        if self.schemas.verbose:
            print('Schema Mapping:')
            print(sorted_schema_mapping_text)
            print()

        self.schemas.paths.schema_mapping_path.parent.mkdir(parents=True, exist_ok=True)
        self.schemas.paths.schema_mapping_path.write_text(sorted_schema_mapping_text)

        # we now check if schemas inside the same topic have the same key
        last_key_list: List[str] = []
        last_topic = ""
        for name, mapping in sorted_schema_mapping.items():
            self.current_schema_name = name
            topic = AvroNaming.get_topic(self.schemas.paths.base_namespace, name)
            key_list = mapping[SMF.partition_fields]
            if topic == last_topic:
                if key_list != last_key_list:
                    raise ValueError(
                        "Key list must be the same inside a topic:\n"
                        "\tcurrent schema: {}, key list: {}\n"
                        "\tlast seen key list in the same topic {}:"
                        .format(name, key_list, last_key_list)
                    )

            last_topic = topic
            last_key_list = key_list

    def find_partition_keys(self, node: Avro.Node) -> None:
        """
        Finds property 'partition-field' inside schemas and saves them in the schema mapping.
        :param node: The node
        """
        if isinstance(node, OrderedDict) and Avro.PartitionKey in node:
            if node[Avro.PartitionKey]:
                self.schema_mapping[self.current_schema_name][SMF.partition_fields] \
                    .append(node[Avro.Name])

            # property 'partition-field' is removed from the schema anyway
            node.pop(Avro.PartitionKey)

    def find_logical_types(self, node: Avro.Node) -> None:
        """
        Finds logicalType 'user-id-fields' inside schemas and saves them in the schema mapping.
        :param node: The node
        """
        if self.ancestors and isinstance(node, OrderedDict):
            if Avro.LogicalType in node and node[Avro.LogicalType] == 'user_id':
                fully_qualified_field_name: List[str] = []

                for ancestor in reversed(self.ancestors):
                    if Avro.Type in ancestor.node \
                            and isinstance(ancestor.node, OrderedDict) \
                            and ancestor.node[Avro.Type] in \
                            self.schemas.paths.schema_mapping_user_id_types_exclude:
                        return

                for ancestor in self.ancestors:
                    if Avro.Name in ancestor.node \
                            and isinstance(ancestor.node, OrderedDict) \
                            and Avro.Type in ancestor.node \
                            and ancestor.node[Avro.Type] != Avro.Record:
                        fully_qualified_field_name.append(ancestor.node[Avro.Name])

                if len(fully_qualified_field_name) >= 1:
                    field_name = '.'.join(fully_qualified_field_name)
                    user_id_fields = \
                        self.schema_mapping[self.current_schema_name][SMF.user_id_fields]
                    if field_name not in user_id_fields:
                        user_id_fields.append(field_name)



#!/usr/bin/env python

"""
Command line entrypoint for avro preprocessor
"""

import argparse
import json
from pathlib import Path

from avro_preprocessor.avro_paths import AvroPaths
from avro_preprocessor.preprocessor import AvroPreprocessor

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

if __name__ == '__main__':
    PARSER = argparse.ArgumentParser(
        description='Pre-processor for extended Avro schemata.')

    PARSER.add_argument('-a', '--avro-tools', dest='avro_tools', type=str, default=None)

    PARSER.add_argument('-i', '--input-path', dest='input_path', type=str, required=True)
    PARSER.add_argument('-o', '--output-path', dest='output_path', type=str, required=True)

    PARSER.add_argument('-b', '--base-namespace', dest='base_namespace', type=str, required=True)

    TYPES = PARSER.add_mutually_exclusive_group(required=True)
    TYPES.add_argument('-t', '--types-namespace', dest='types_namespace', type=str)
    TYPES.add_argument(
        '-n', '--no-types-namespace', dest='types_namespace', action='store_const', const=None)

    RPC = PARSER.add_mutually_exclusive_group(required=True)
    RPC.add_argument('-r', '--rpc-namespace', dest='rpc_namespace', type=str)
    RPC.add_argument(
        '-e', '--no-rpc-namespace', dest='rpc_namespace', action='store_const', const=None)

    PARSER.add_argument('-d', '--metadata-schema', dest='metadata_schema', type=str, default=None)
    PARSER.add_argument(
        '-de', '--metadata-exclude', dest='metadata_exclude', type=str, default=None)

    PARSER.add_argument('-k', '--key-schema', dest='key_schema', type=str, default=None)

    PARSER.add_argument(
        '-ksns',
        '--key-subject-name-strategy',
        dest='key_subject_name_strategy',
        type=str,
        default='TopicRecordNameStrategy'
    )

    PARSER.add_argument(
        '-p', '--schema-mapping-path', dest='schema_mapping_path', type=str, default=None)
    PARSER.add_argument(
        '-pe',
        '--exclude-from-schema-mapping',
        dest='exclude_from_schema_mapping',
        type=str,
        default=None
    )
    PARSER.add_argument(
        '-pet',
        '--user-id-types-exclude-from-schema-mapping',
        dest='user_id_types_exclude_from_schema_mapping',
        type=str,
        default=None
    )

    PARSER.add_argument(
        '-ie', '--input-schema-file-extension', dest='input_schema_file_extension', type=str,
        default='exavsc', help="Comma separated list of extensions, e.g. 'exavsc,yexavsc'")

    PARSER.add_argument(
        '-oe', '--output-schema-file-extension', dest='output_schema_file_extension', type=str,
        default='avsc')

    PARSER.add_argument(
        '-if', '--input-schema-file-format', dest='input_schema_file_format', type=str,
        default='json', help="Can be json or yaml")

    PARSER.add_argument(
        '-of', '--output-schema-file-format', dest='output_schema_file_format', type=str,
        default='json', help="Can be json or yaml")

    PARSER.add_argument('-v', '--verbose', dest='verbose', action='store_true')
    PARSER.set_defaults(verbose=False)

    PARSER.add_argument('-s', '--schema-registry', dest='schema_registry', type=str, default='')

    INDENT = PARSER.add_mutually_exclusive_group()
    INDENT.add_argument('-j', '--json_indent', dest='json_indent', type=int, default=4)
    INDENT.add_argument(
        '-c', '--json_compact', dest='json_indent', action='store_const', const=None)

    PARSER.add_argument('-y', '--yaml_indent', dest='yaml_indent', type=json.loads,
                        default='{"mapping":2,"sequence":4,"offset":2}',
                        help="yaml indentation (as json), see "
                             "https://yaml.readthedocs.io/en/latest/"
                             "detail.html?highlight=indent#indentation-of-block-sequences "
                             """use {"mapping": 2, "sequence": 2, "offset": 0} """ 
                             "for avoid indenting sequence values")

    AVAILABLE_MODULES = ' '.join(AvroPreprocessor.available_preprocessing_modules.keys())
    PARSER.add_argument('-m', '--modules', dest='modules', nargs='*', default=None,
                        help='Available modules: {}'.format(AVAILABLE_MODULES))

    ARGS = PARSER.parse_args()

    SCHEMA_MAPPING_PATH = Path(ARGS.schema_mapping_path).absolute() if ARGS.schema_mapping_path \
        else Path('./schema-mapping.json')

    METADATA_EXCLUDE = ARGS.metadata_exclude.split(',') if ARGS.metadata_exclude else None

    SCHEMA_MAPPING_EXCLUDE = ARGS.exclude_from_schema_mapping \
        .split(',') if ARGS.exclude_from_schema_mapping else None

    SCHEMA_MAPPING_USER_ID_TYPES_EXCLUDE = ARGS.user_id_types_exclude_from_schema_mapping \
        .split(',') if ARGS.user_id_types_exclude_from_schema_mapping else None

    AVRO_PREPROCESSOR: AvroPreprocessor = AvroPreprocessor(
        AvroPaths(
            input_path=ARGS.input_path,
            output_path=ARGS.output_path,
            base_namespace=ARGS.base_namespace,
            types_namespace=ARGS.types_namespace,
            rpc_namespace=ARGS.rpc_namespace,
            metadata_schema=ARGS.metadata_schema,
            metadata_exclude=METADATA_EXCLUDE,
            key_schema=ARGS.key_schema,
            key_subject_name_strategy=ARGS.key_subject_name_strategy,
            input_schema_file_extension=ARGS.input_schema_file_extension,
            output_schema_file_extension=ARGS.output_schema_file_extension,
            input_schema_file_format=ARGS.input_schema_file_format,
            output_schema_file_format=ARGS.output_schema_file_format,
            schema_mapping_path=SCHEMA_MAPPING_PATH,
            schema_mapping_exclude=SCHEMA_MAPPING_EXCLUDE,
            schema_mapping_user_id_types_exclude=SCHEMA_MAPPING_USER_ID_TYPES_EXCLUDE,
            avro_tools_path=ARGS.avro_tools
        ),
        verbose=ARGS.verbose,
        json_indent=ARGS.json_indent,
        yaml_indent=ARGS.yaml_indent,
        schema_registry_url=ARGS.schema_registry
    )

    AVRO_PREPROCESSOR.process(ARGS.modules)



"""
# Support to use a list for lists as "doc" field to deal with JSON obnoxious lack of
# newline support in strings
"""
from collections import OrderedDict

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class DocumentationCondenser(PreprocessorModule):
    """
    Condense documentation fields ('doc').
    """

    def process(self) -> None:
        """Process all schemas."""

        # This module has to be applied even to the original schemas
        for _, schema in self.original_schemas_iter():
            self.traverse_schema(self.condense_documentation, schema)

        for _, schema in self.processed_schemas_iter():
            self.traverse_schema(self.condense_documentation, schema)

    @staticmethod
    def condense_documentation(node: Avro.Node) -> None:
        """
        Condense the "doc" field from list to one string

        :param node: The (sub) node
        """

        if isinstance(node, OrderedDict):
            if Avro.Doc in node:
                if isinstance(node[Avro.Doc], list):
                    node[Avro.Doc] = ' '.join(node[Avro.Doc])



"""
A module to add a standard metadata record to every event.
"""
import copy
from collections import OrderedDict

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class MetadataAdder(PreprocessorModule):
    """
    Adds the metadata type to every event.
    """

    def process(self) -> None:
        """Process all schemas."""

        # If types namespace was defined by the user
        if self.schemas.paths.types_namespace:

            # Let's check if the metadata type exists
            metadata_fully_qualified_name = self.schemas.paths.metadata_schema
            if metadata_fully_qualified_name in self.schemas.original:

                metadata = self.schemas.original[metadata_fully_qualified_name]
                metadata_field = OrderedDict((
                    (Avro.Name, "metadata"),
                    (Avro.Doc, 'Standard metadata for all schemas'),
                    (Avro.Type, metadata)
                ))

                for name, schema in self.processed_schemas_iter():

                    # metadata added only to domain schemas (that are not excluded).
                    # Added as first element in the fields list
                    if self.schemas.paths.is_event_resource(name) and all(
                            not name.startswith(exclude)
                            for exclude in self.schemas.paths.metadata_exclude
                    ):
                        schema[Avro.Fields].insert(0, copy.deepcopy(metadata_field))



"""
Class to define input and output paths for Avro schema directory trees.
"""

import glob
import shutil
from pathlib import Path
from typing import Callable, Optional, List

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class AvroPaths:  # pylint: disable=R0902
    """Class for keeping track of Avro schemas path."""

    def __init__(
            self,
            input_path: str,
            output_path: str,
            base_namespace: str,
            types_namespace: Optional[str] = None,
            rpc_namespace: Optional[str] = None,
            metadata_schema: Optional[str] = None,
            metadata_exclude: Optional[List[str]] = None,
            key_schema: Optional[str] = None,
            key_subject_name_strategy: str = "TopicRecordNameStrategy",
            input_schema_file_extension: str = 'exavsc',
            output_schema_file_extension: str = 'avsc',
            input_schema_file_format: str = 'json',
            output_schema_file_format: str = 'json',
            schema_mapping_path: Path = Path('./schema-mapping.json'),
            schema_mapping_exclude: Optional[List[str]] = None,
            schema_mapping_user_id_types_exclude: Optional[List[str]] = None,
            avro_tools_path: Optional[str] = None
    ):  # pylint: disable=R0913
        """
        :param input_path: The root input path
        :param output_path: The root output path
        :param base_namespace: The root namespace for schemas (e.g. 'com.example')
        :param types_namespace: The namespace of types (reusable schemas not registered to a topic)
        :param rpc_namespace: The namespace of rpc protocols
        :param metadata_schema: The fully qualified name of the metadata type schema
        :param key_schema: The fully qualified name of the key type schema
        :param input_schema_file_extension: The filename extension of input Avro files
        :param output_schema_file_extension: The filename extension of output Avro files
        :param input_schema_file_format: The format of input Avro files
        :param output_schema_file_format: The format of output Avro files
        :param schema_mapping_path: The path of the schema mapping JSON file
        :param schema_mapping_exclude: The paths excluded from the schema mapping
        :param schema_mapping_user_id_types_exclude: The containers (e.g. array or map) to exclude
        :param avro_tools_path: The path to the Avro tools jar
        """
        self.input_path: str = input_path
        self.output_path: str = output_path
        self.base_namespace: str = base_namespace
        self.types_namespace: Optional[str] = types_namespace
        if self.types_namespace and not self.types_namespace.startswith(self.base_namespace):
            raise ValueError("Types namespace {} should be part of base namespace {}."
                             .format(self.types_namespace, self.base_namespace))
        self.rpc_namespace: Optional[str] = rpc_namespace
        if self.rpc_namespace and not self.rpc_namespace.startswith(self.base_namespace):
            raise ValueError("RPC namespace {} should be part of base namespace {}."
                             .format(self.rpc_namespace, self.base_namespace))
        self.metadata_schema: Optional[str] = metadata_schema
        self.metadata_exclude: List[str] = metadata_exclude if metadata_exclude else []
        self.key_schema: Optional[str] = key_schema
        self.key_subject_name_strategy: str = key_subject_name_strategy
        self.input_schema_file_extension: List[str] = \
            input_schema_file_extension.replace('.', '').split(',')
        self.output_schema_file_extension: str = output_schema_file_extension.replace('.', '')
        self.input_schema_file_format = input_schema_file_format
        self.output_schema_file_format = output_schema_file_format
        self.schema_mapping_path: Path = schema_mapping_path
        self.schema_mapping_exclude: List[str] = \
            schema_mapping_exclude if schema_mapping_exclude else []
        self.schema_mapping_user_id_types_exclude: List[str] = \
            schema_mapping_user_id_types_exclude if schema_mapping_user_id_types_exclude else []
        self.avro_tools_path: Optional[str] = avro_tools_path

    def to_output_path(self, name: str) -> Path:
        """
        Schema fully qualified name to output path
        :param name: schema name
        :return: the output path
        """
        return Path(self.output_path).joinpath(Path(name.replace('.', '/'))) \
            .with_suffix('.' + self.output_schema_file_extension)

    def to_input_path(self, name: str, extension: str) -> Path:
        """
        Schema fully qualified name to input path
        :param name: schema name
        :param extension: schema extension
        :return: the input path
        """
        return Path(self.input_path).joinpath(Path(name.replace('.', '/'))) \
            .with_suffix('.' + extension)

    def is_type_resource(self, schema_namespace: str) -> bool:
        """
        Returns True if a schema namespace is type, i.e. it is a reusable Avro resource
        not to be registered to a Kafka topic.
        :param schema_namespace: The namespace of this schema
        :return: bool
        """
        if self.types_namespace is not None and schema_namespace.startswith(self.types_namespace):
            return True

        return False

    def is_rpc_resource(self, schema_namespace: str) -> bool:
        """
        Returns True if a schema namespace is a rpc resource, i.e. it is an Avro resource
        for RPC protocol communications.
        :param schema_namespace: The namespace of this schema
        :return: bool
        """
        if self.rpc_namespace is not None and schema_namespace.startswith(self.rpc_namespace):
            return True

        return False

    def is_event_resource(self, schema_namespace: str) -> bool:
        """
        Returns True if a schema namespace is an event schema resource.
        :param schema_namespace: The namespace of this schema
        :return: bool
        """
        if self.is_type_resource(schema_namespace) or self.is_rpc_resource(schema_namespace):
            return False

        return True

    @staticmethod
    def traverse_path(root_path: str, file_extension: List[str], func: Callable[[Path], None]) \
            -> None:
        """
        Traverse a path and applies 'func' to each file matching the extension in input.
        :param root_path: The root path
        :param file_extension: The extension to match
        :param func: The function to apply
        """
        for file in glob.iglob(root_path + '/**/*', recursive=True):
            path = Path(file).resolve()
            if path.suffix.replace('.', '') in file_extension:
                func(path)

    @staticmethod
    def reset_directory(directory: str) -> None:
        """
        Deletes and re-creates a directory
        :param directory: The directory
        """
        shutil.rmtree(directory, ignore_errors=True)
        Path(directory).parent.mkdir(parents=True, exist_ok=True)



"""
A module to register Avro schemas to Confluent Schema Registry.
It uses the fully qualified name of a schema according to the strategy
"io.confluent.kafka.serializers.subject.TopicRecordNameStrategy".
It is used to employ multiple schemas per topic.
See [https://www.confluent.io/blog/put-several-event-types-kafka-topic/]
"""

import json

import requests
from requests import Response

from avro_preprocessor.modules.schema_registrar import SchemaRegistrar

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class SchemaRegistryChecker(SchemaRegistrar):
    """
    Checks the compatibility of a schema on the schema registry.
    """

    def get_request_url(self, subject: str) -> str:
        """
        Returns the URL for this request
        :param subject: The Avro subject name
        :return: The full URL
        """
        url = self.schema_registry_url + "/compatibility/subjects/" + subject + "/versions/latest"
        return url

    def process_response(self, response: Response) -> None:
        """
        Process the response.
        :param response: The response
        """

        if response.status_code not in (requests.codes.ok, 404):  # pylint: disable=E1101
            print(response)
            response.raise_for_status()

        result = json.loads(response.content)
        if self.schemas.verbose:
            print('Response:', result, '\n')

        if 'is_compatible' in result and not result['is_compatible']:
            raise ValueError("Schema Registry compatibility failed for \n\tschema: {} \n\turl: {}"
                             .format(self.current_schema_name, self.url))



"""
Module providing a custom order for json-based Avro schemas.
"""
from collections import OrderedDict
from typing import Union

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

from dataclasses import dataclass


@dataclass
class Avro:  # pylint: disable-all
    """
    Helper class to avoid plain strings as dictionary keys.
    """
    Name = 'name'
    Namespace = 'namespace'
    Type = 'type'
    Fields = 'fields'
    Items = 'items'
    Doc = 'doc'
    Record = 'record'
    Error = 'error'
    Enum = 'enum'
    Default = 'default'
    LogicalType = 'logicalType'
    Symbols = 'symbols'
    Array = 'array'
    Map = 'map'
    Aliases = 'aliases'
    Order = 'order'
    Values = 'values'
    Size = 'size'
    Precision = 'precision'
    Scale = 'scale'

    Protocol = 'protocol'
    Messages = 'messages'
    Request = 'request'
    Response = 'response'
    Types = 'types'
    Errors = 'errors'
    OneWay = 'one-way'

    Optional = 'nullable_optional'
    PartitionKey = 'partition-field'

    Null = 'null'
    Value = 'value'

    Node = Union[str, list, OrderedDict]
    Key = Union[slice, int, None]



"""
Container for Avro schemas.
"""

import copy
import json

from collections import OrderedDict
from json import JSONDecodeError
from pathlib import Path
from typing import Optional, Union, cast
from ruamel.yaml import YAML  # type: ignore

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.avro_naming import AvroNaming
from avro_preprocessor.avro_paths import AvroPaths

from avro_preprocessor.colored_json import ColoredJson

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class SchemasContainer:
    """
    A memory container from schemas.
    """

    def __init__(
            self,
            paths: AvroPaths,
            verbose: bool = True,
    ) -> None:
        self.paths: AvroPaths = paths
        self.verbose: bool = verbose
        self.original: OrderedDict = OrderedDict()
        self.processed: OrderedDict = OrderedDict()
        self.autogenerated_keys: OrderedDict = OrderedDict()

        self.modules: OrderedDict = OrderedDict()

        self.original_string: OrderedDict = OrderedDict()

    def read_schemas(self) -> None:
        """
        Reads schemas from disk
        """
        self.paths.traverse_path(
            self.paths.input_path, self.paths.input_schema_file_extension, self._get_schema)

        self.processed = copy.deepcopy(self.original)

        if self.verbose:
            print('')

    def _get_schema(self, path: Path) -> None:
        schema_text = path.read_text()
        try:
            try:
                schema = json.loads(schema_text, object_pairs_hook=OrderedDict)
            except JSONDecodeError:
                yaml = YAML()
                schema = yaml.load(schema_text)
                # print(ColoredJson.highlight_yaml(schema))

        except Exception as e:
            import sys
            raise type(e)(str(e) + '\nError while processing schema %s' % path) \
                .with_traceback(sys.exc_info()[2])
        namespace = str(
            Path(
                str(path.absolute())
                .replace(str(Path(self.paths.input_path).resolve()) + '/', '')
                .replace('/', '.')
            )
            .with_suffix('')
        )

        self._store_schema(schema, namespace)
        self.original_string[namespace] = schema_text

        if self.verbose:
            print('Reading', path, ' -> ', namespace)

    def _store_schema(self, schema: OrderedDict, namespace: str) -> None:
        schema_doc = schema[Avro.Doc] if isinstance(schema[Avro.Doc], str) \
            else ' '.join(schema[Avro.Doc])

        # if we run the preprocessor with the KeyGenerator module and then re-run it again
        # we need to recognize the autogenerated key schemas and store them
        # in the right container
        if schema_doc.startswith(AvroNaming.autogenerated_key_doc):
            self.autogenerated_keys[namespace] = schema
        else:
            self.original[namespace] = schema

    def write_schemas(self) -> None:
        """
        Writes schemas to disk.
        """
        # let's cleanup the output schema directory first
        AvroPaths.reset_directory(self.paths.output_path)

        schemas_and_keys = list(self.processed.items()) + list(self.autogenerated_keys.items())

        for name, schema in schemas_and_keys:
            path = self.paths.to_output_path(name)
            path.parent.mkdir(parents=True, exist_ok=True)

            if self.paths.output_schema_file_format.lower() == 'yaml':
                yaml = YAML()
                yaml.default_flow_style = False
                yaml.indent(**ColoredJson.yaml_indent)
                yaml.Representer.add_representer(OrderedDict, yaml.Representer.represent_dict)
                yaml.dump(schema, path)
            else:
                schema_text = json.dumps(schema, indent=ColoredJson.json_indent)
                path.write_text(schema_text)

    def print_schemas(self, how: str = 'json', indent: Union[None, int, dict] = None) -> None:
        """
        Print all schemas.
        :param how: 'json' or 'yaml' or 'dict'
        :param indent: how much to indent printed JSONs or YAMLs
        """

        def print_schema(schema_name: str, data: OrderedDict) -> None:
            value: Union[str, OrderedDict]
            if how == 'json':
                value = ColoredJson.highlight_json(data, cast(Optional[int], indent))
            elif how == 'yaml':
                value = ColoredJson.highlight_yaml(data, cast(Optional[dict], indent))
            else:  # 'dict'
                value = data
            print(schema_name, value)

        print('#### ORIGINAL SCHEMAS ####')
        for name, schema in self.original.items():
            print_schema(name, schema)

        print('#### PROCESSED SCHEMAS ####')
        for name, schema in self.processed.items():
            print_schema(name, schema)



"""
A module to create java classes from Avro schemas.
"""
import glob
import subprocess
from pathlib import Path
from typing import Tuple, List, cast

from avro_preprocessor.avro_paths import AvroPaths
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class JavaClassesCreator(PreprocessorModule):
    """
    Create Java classes for all schemas.
    """
    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)
        self.java_classes_dir = Path(self.schemas.paths.input_path).parent.joinpath('java_classes')

    def process(self) -> None:
        """Process all schemas."""
        self.delete_java_classes()
        self.schemas.write_schemas()

        for name, _ in self.processed_schemas_and_keys_iter():
            if not self.schemas.paths.is_type_resource(name):
                schema_path = self.schemas.paths.to_output_path(name)
                ret_code, _, _ = self.schema_to_java(
                    schema_path, self.java_classes_dir,
                    'protocol' if self.schemas.paths.is_rpc_resource(name) else 'schema'
                )
                if ret_code != 0:
                    raise ValueError("The Java classes creation failed.")

    def schema_to_java(self,
                       schema_path: Path,
                       destination_path: Path,
                       schema_type: str) -> Tuple[int, str, str]:
        """
        Compile Java classes with Avro tools and check if the work correctly.
        :param schema_path: The path of the schema in input
        :param destination_path: The destination path to output the Java classes
        :param schema_type: 'schema' or 'protocol'
        :return: Returns the return code of avro-tools
        """

        command = [
            'java',
            '-jar',
            str(Path(cast(str, self.schemas.paths.avro_tools_path)).absolute()),
            'compile',
            schema_type,
            str(schema_path),
            str(destination_path)
        ]

        if self.schemas.verbose:
            print(' '.join(command))

        process = subprocess.run(command, capture_output=True, text=True)

        if process.returncode != 0:
            print(process.stdout)
            print(process.stderr)
            print(process.returncode)
        return process.returncode, process.stdout, process.stderr

    def get_java_classes_names(self) -> List[str]:
        """
        Checks that the generated Java classes have the expected names.
        :return: The names found.
        """
        found_classes = set(
            str(Path(fname).with_suffix('').name)
            for fname in glob.glob(str(self.java_classes_dir) + '/**/*.java', recursive=True)
        )
        return sorted(found_classes)

    def delete_java_classes(self) -> None:
        """
        Delete generated java classes.
        """
        # let's cleanup the classes directory first
        AvroPaths.reset_directory(str(self.java_classes_dir))



"""
A module to expand references of types in schemas using a Depth First strategy.
"""

import copy
from collections import OrderedDict
from typing import Set, Union, cast

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class ReferencesExpander(PreprocessorModule):
    """
    Expands References in all schemas.
    """
    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)

        self.per_schema_already_expanded_records: Set = set()

    def process(self) -> None:
        """Process all schemas."""

        for _, schema in self.processed_schemas_iter():
            self.per_schema_already_expanded_records = set()
            self.traverse_schema(self.expand_ref_in_record, schema)

    def expand_ref_in_record(self, node: Avro.Node) -> bool:
        """
        Expands references in an Avro Node.
        :param node: The node
        """
        if self.ancestors and isinstance(node, str):
            original_value = cast(OrderedDict, self.ancestors[-1].node)[self.ancestors[-1].key]
            modified_value = self.do_expand(node)

            if modified_value != original_value:
                cast(OrderedDict, self.ancestors[-1].node)[self.ancestors[-1].key] = modified_value
                return True

        return False

    def do_expand(self, resource_to_expand: str) -> Union[str, OrderedDict]:
        """
        Here we do_expand a resource (e.g. com.example.event.type.Gender) if it complies with
        certain criteria (they are in the "if").

        :param resource_to_expand: the resource
        :return: Whether the criteria were met or not
        """
        if '.' in resource_to_expand \
                and resource_to_expand not in self.per_schema_already_expanded_records \
                and resource_to_expand in self.schemas.original:

            self.per_schema_already_expanded_records.add(resource_to_expand)
            return cast(OrderedDict, copy.deepcopy(self.schemas.original[resource_to_expand]))

        return resource_to_expand



"""
Definition of Avro topics and subjects given a schema name.

It uses the fully qualified name of a schema according to the strategy
"io.confluent.kafka.serializers.subject.TopicRecordNameStrategy".
It is used to employ multiple schemas per topic.
See [https://www.confluent.io/blog/put-several-event-types-kafka-topic/]
"""

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class AvroNaming:
    """
    Returns topic and subject names given a schema fully qualified name
    """

    autogenerated_key_doc = "The autogenerated Kafka key for topic"

    @classmethod
    def get_subject_name_for_value(
            cls, base_namespace: str, schema_fully_qualified_name: str) -> str:
        """
        Returns the Avro subject name given the schema fully qualified name.

        The fully qualified name of a schema is generated according to
        "io.confluent.kafka.serializers.subject.TopicRecordNameStrategy"

        :param base_namespace: The schemas base namespace
        :param schema_fully_qualified_name: Fully qualified name of the schema
        :return: The subject name
        """
        topic = cls.get_topic(base_namespace, schema_fully_qualified_name)
        return topic + '-' + schema_fully_qualified_name

    @classmethod
    def get_subject_name_for_key(
            cls, base_namespace: str, schema_fully_qualified_name: str) -> str:
        """
        Returns the Avro subject name given the schema fully qualified name.

        The fully qualified name of a schema is generated according to
        "io.confluent.kafka.serializers.subject.TopicRecordNameStrategy"

        # It uses the "io.confluent.kafka.serializers.subject.TopicNameStrategy",
        # i.e. "<topic>-key"

        :param base_namespace: The schemas base namespace
        :param schema_fully_qualified_name: Fully qualified name of the schema
        :return: The subject name
        """
        topic = cls.get_topic(base_namespace, schema_fully_qualified_name)
        return topic + '-' + schema_fully_qualified_name
        # return topic + '-key'

    @classmethod
    def get_topic(cls, base_namespace: str, schema_fully_qualified_name: str) -> str:
        """
        Get the topic name given the schema.

        Topic name is the last two parents before schema name, e.g.

        'com.jaumo.event.domain.user.users.RoleAssigned' -> 'user.users'

        :param base_namespace: The schemas base namespace
        :param schema_fully_qualified_name: Fully qualified name of the schema
        :return: The topic name
        """
        return '.'.join(
            schema_fully_qualified_name.replace(base_namespace + '.', '').split('.')[:-1]
        )

    @classmethod
    def get_key_fully_qualified_name(cls, schema_fully_qualified_name: str) -> str:
        """
        Get the key name given the schema name.

        :param schema_fully_qualified_name: Fully qualified name of the schema
        :return: The key fully qualified name
        """
        return '.'.join(
            schema_fully_qualified_name.split('.')[:-1] + ['Key']
        )



"""
A module to sorts all the fields of Avro schemas.
"""
import json
from collections import OrderedDict

from avro_preprocessor.colored_json import ColoredJson
from avro_preprocessor.preprocessor_module import PreprocessorModule

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class AvroSorter(PreprocessorModule):
    """
    Sort Avro schemas.
    """

    AVRO_ORDER = [
        'namespace',
        'protocol',  # used in protocols
        'name',
        'doc',
        'aliases',
        'partition-field',  # custom extension
        'nullable_optional',  # custom extension
        'default',
        'type',
        'order',
        'fields',
        'symbols',
        'items',
        'values',
        'size',

        'logicalType',  # used in logical types
        'precision',  # used in logical types
        'scale',  # used in logical types

        'types',  # used in protocols
        'messages',  # used in protocols
        'request',  # used in protocols
        'response',  # used in protocols
        'errors',  # used in protocols
        'one-way',  # used in protocols
    ]

    ORDER = {key: i for i, key in enumerate(AVRO_ORDER)}

    def process(self) -> None:
        """Process all schemas."""
        for name, schema in self.processed_schemas_iter():
            self.schemas.processed[name] = self.sort_avro(schema)

    @classmethod
    def sort_avro(cls, schema: OrderedDict) -> OrderedDict:
        """
        Performs the sort.

        :param schema: An Avro schema
        :return: The sorted schema
        """
        for key, value in schema.items():
            if isinstance(value, OrderedDict):
                schema[key] = cls.sort_avro(value)
            elif isinstance(value, list):
                schema[key] = [cls.sort_avro(element) if not isinstance(element, str) else element
                               for element in value]

        return OrderedDict(sorted(schema.items(), key=lambda item: cls.ORDER.get(item[0], 1000)))


class AvroOrderChecker(AvroSorter):
    """
    Checks if input Avro schemas are correctly sorted.
    """
    def process(self) -> None:
        """Process all schemas."""

        wrong_order = False
        for name, schema in self.processed_schemas_iter():
            sorted_schema = self.sort_avro(schema)
            if schema != sorted_schema:
                wrong_order = True
                print('########## Schema', name, 'is sorted incorrectly:')
                print(json.dumps(schema, indent=ColoredJson.json_indent))
                print('\n########## Should be:')
                print(json.dumps(sorted_schema, indent=ColoredJson.json_indent))
                print('\n\n')

        if wrong_order:
            raise ValueError(
                "Schema properties should be sorted according to the following order:\n\n\t{}"
                "\n\nSchemas with correct order where printed above."
                .format('\n\t'.join(self.AVRO_ORDER))
            )



"""
A module to expand nullable_optional fields.
"""

from collections import OrderedDict
from typing import Set, Optional, Tuple, cast

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class NullableOptionalExpander(PreprocessorModule):
    """
    Expand nullable_optional fields in all schemas.
    """
    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)

        self.per_schema_nullable_optional_records: Set = set()
        self.current_namespace: Optional[str] = None

    def process(self) -> None:
        """Process all schemas."""
        for _, schema in self.processed_schemas_iter():
            # we keep track of the namespace; used to refer to already substituted custom records
            if Avro.Namespace in schema:
                self.current_namespace = schema[Avro.Namespace]
            self.per_schema_nullable_optional_records = set()
            self.traverse_schema(self.substitute_field_type, schema)

    def get_subrecord_name(self, record: OrderedDict) -> Tuple[str, str]:
        """
        Generates the name of the subrecord for nullable optionals

        :param record: The record in input
        :return: The new name and its fully qualified version
        """
        def capitalize_only_first(word: str) -> str:
            return word[0].upper() + word[1:]

        if isinstance(record[Avro.Type], str):
            if '.' in record[Avro.Type]:
                tokens = record[Avro.Type].split('.')
                new_record_name = \
                    '.'.join(
                        tokens[:-1] + ['Optional' + capitalize_only_first(tokens[-1])]
                    )
            else:
                new_record_name = 'Optional' + capitalize_only_first(record[Avro.Type])
        else:
            if Avro.Name in record[Avro.Type] and Avro.Namespace in record[Avro.Type]:
                new_record_name = '.'.join((
                    record[Avro.Type][Avro.Namespace],
                    'Optional' + capitalize_only_first(record[Avro.Type][Avro.Name])
                ))
            else:
                new_record_name = 'Optional' + capitalize_only_first(record[Avro.Name])

        if self.current_namespace:
            if self.schemas.paths.base_namespace not in new_record_name:
                fully_qualified_subrecord_name = \
                    self.current_namespace + '.' + new_record_name
            else:
                fully_qualified_subrecord_name = new_record_name
        else:
            raise ValueError("self.current_namespace was not initialized for"
                             "name {}, type{}".format(record[Avro.Name], record[Avro.Type]))

        return new_record_name, fully_qualified_subrecord_name

    def substitute_field_type(self, record: Avro.Node) -> None:
        """
        Substitutes in place a nullable_optional record with with Optional<Type>
        (e.g. string -> OptionalString).

        Substitutes in place a nullable_optional record with with Optional<Name>
        (e.g. name = address, type = record -> OptionalAddress).
        :param record: The field in input
        """
        if isinstance(record, OrderedDict) and Avro.Optional in record:
            if record[Avro.Optional] is True:

                subrecord_name, fully_qualified_subrecord_name = self.get_subrecord_name(record)

                if fully_qualified_subrecord_name in self.per_schema_nullable_optional_records:

                    # We cannot define the same record twice in the same file,
                    # so we use a reference here
                    record[Avro.Type] = [Avro.Null, fully_qualified_subrecord_name]
                else:

                    # Saving the reference
                    self.per_schema_nullable_optional_records.add(fully_qualified_subrecord_name)

                    def add_null_type() -> list:
                        current_type = record[cast(slice, Avro.Type)]  # cast to make mypy happy
                        if isinstance(current_type, list):
                            # avro type is a union
                            if Avro.Null in current_type:
                                # "null" already in the list
                                return current_type
                            # we add "null" in the first position of the list
                            return [Avro.Null] + current_type
                        return [Avro.Null, current_type]

                    # Substituting the nullable_optional record with a wrapper:
                    # Semantics:
                    # if MyField == null -> not present
                    # if MyField.value == null -> set-to-null
                    # if MyField.value != null -> set-to-value
                    record[Avro.Type] = [
                        Avro.Null,
                        OrderedDict((
                            (Avro.Name, subrecord_name),
                            (Avro.Type, Avro.Record),
                            (Avro.Fields, [
                                OrderedDict((
                                    (Avro.Name, Avro.Value),
                                    (Avro.Doc, "The optional value"),
                                    (Avro.Type, add_null_type())
                                ))
                            ]),
                        ))
                    ]

                    subrecord: OrderedDict = record[Avro.Type][1]

                    if Avro.Doc in record:
                        subrecord[Avro.Doc] = record[Avro.Doc] + ' (Optional Value)'

                    # Avro complex types have additional fields that must be moved to the subrecord
                    additional_fields = \
                        [(field, record[field]) for field in record
                         if field not in
                         (Avro.Name, Avro.Type, Avro.Doc, Avro.Default, Avro.Optional)]
                    for key, value in additional_fields:
                        subrecord[key] = value
                        record.pop(key)

                # necessary to allow 'not present' semantics
                record[Avro.Default] = None

            # property 'nullable_optional' is removed from the schema anyway
            record.pop(Avro.Optional)



#!/usr/bin/env python3
"""
Test class for the Avro schema extension.
"""
import json
import os
import unittest
from collections import OrderedDict
from pathlib import Path
from typing import Set, Any, Dict

from avro_preprocessor.avro_paths import AvroPaths
from avro_preprocessor.modules.java_classes_creator import JavaClassesCreator
# from avro_preprocessor.modules.schema_registrar import SchemaRegistrar
from avro_preprocessor.preprocessor import AvroPreprocessor
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"

ROOT_DIR = Path(__file__).absolute().parent.parent
FIXTURES_DIR = ROOT_DIR.joinpath('fixtures/')
JAVA_CLASSES_DIR = ROOT_DIR.joinpath('java_classes/')
AVRO_TOOLS_JAR = str(ROOT_DIR.joinpath('avro-tools-1.9.0.jar'))


class AvroPreprocessorTest(unittest.TestCase):
    """
    Test class for the Avro schema extension.
    """

    try:
        schema_registry_url = os.environ['SCHEMA_REGISTRY']
    except KeyError as e:
        schema_registry_url = 'http://localhost:8081'

    SUCCESS = 0

    schema_header = """
    {
        "namespace": "com.jaumo.event.domain.user.users",
        "name": "TestRecord",
        "doc": "Test extended record",
        "type": "record",
        "fields": [
    """

    schema_footer = """
        ]
    }
    """

    @unittest.skip("fixture not available")
    def test_large_set(self) -> None:
        """
        Test on a large set of schemas.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('../../event_schema/schema/')),
                output_path=str(FIXTURES_DIR.joinpath('../../event_schema/build/schema/')),
                base_namespace='com.jaumo.message_schema',
                types_namespace='com.jaumo.message_schema.type',
                rpc_namespace='com.jaumo.message_schema.rpc',
                metadata_schema='com.jaumo.message_schema.type.domain.metadata.Metadata',
                metadata_exclude=['com.jaumo.message_schema.simple_rpc'],
                key_schema='com.jaumo.message_schema.type.domain.key.DefaultKey',
                key_subject_name_strategy='RecordNameStrategy',
                input_schema_file_extension='exavsc',
                schema_mapping_path=
                ROOT_DIR.joinpath('../event_schema/build/schema-mapping.json'),
                schema_mapping_exclude=[
                    'com.jaumo.message_schema.simple_rpc', 'com.jaumo.message_schema.mqtt'],
                schema_mapping_user_id_types_exclude=['array', 'map'],
                avro_tools_path=AVRO_TOOLS_JAR,
            ),
            verbose=True,
            json_indent=4
        )

        # avro_preprocessor.process()
        import copy
        avro_preprocessor.process([
            m.__name__ for m in
            copy.deepcopy(AvroPreprocessor.preprocessing_modules)
            if m not in [JavaClassesCreator]
            # if m not in [SchemaRegistrar, JavaClassesCreator]
        ])

        # self.assert_trees_equals(
        #     '../../event_schema/build/schema/', '../../event_schema/build/schema_expected/')

    def test_full_t1(self) -> None:
        """
        Full test on fixture 't1_input'.
        """
        paths = AvroPaths(
            input_path=str(FIXTURES_DIR.joinpath('t1_input/')),
            output_path=str(FIXTURES_DIR.joinpath('t1_output/')),
            base_namespace='com.jaumo.event',
            types_namespace='com.jaumo.event.type',
            metadata_schema='com.jaumo.event.type.Metadata',
            input_schema_file_extension='exavsc',
            schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
            avro_tools_path=AVRO_TOOLS_JAR
        )

        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths, verbose=True, json_indent=4, schema_registry_url=self.schema_registry_url)
        avro_preprocessor.process()

        self.assert_trees_equals('t1_expected/', 't1_output/')

        self.assertEqual(
            JavaClassesCreator(avro_preprocessor.schemas).get_java_classes_names(),
            sorted({
                'Address',
                'OptionalAddress',
                'OptionalInt_or_string',
                'OptionalString',
                'TestRecursive',
                'Kind',
                'Letter',
                'OptionalInt',
                'OptionalKind',
                'OptionalLetter',
                'OptionalString',
                'UserUpdate'
            }))

    def test_full_t2(self) -> None:
        """
        Full test on fixture 't2_input_unsorted'.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('t2_input/')),
                output_path=str(FIXTURES_DIR.joinpath('t2_output/')),
                base_namespace='com.jaumo.event',
                types_namespace='com.jaumo.event.type',
                metadata_schema='com.jaumo.event.type.Metadata',
                input_schema_file_extension='exavsc',
                schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
                avro_tools_path=AVRO_TOOLS_JAR
            ),
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process()

        self.assert_trees_equals('t2_expected/', 't2_output/')

    def test_full_t3(self) -> None:
        """
        Full test on fixture 't3_input'.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('t3_input/')),
                output_path=str(FIXTURES_DIR.joinpath('t3_output/')),
                base_namespace='com.jaumo.schema',
                types_namespace='com.jaumo.schema.type',
                rpc_namespace='com.jaumo.schema.rpc',
                metadata_schema='com.jaumo.schema.type.Metadata',
                input_schema_file_extension='exavsc',
                schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
                avro_tools_path=AVRO_TOOLS_JAR
            ),
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process()

        self.assert_trees_equals('t3_expected/', 't3_output/')

    def test_full_t3_yaml(self) -> None:
        """
        Full test on fixture 't3_input' but converting to yaml first.
        """
        preproc_args = {
            'verbose': True,
            'schema_registry_url': self.schema_registry_url,
        }

        paths_args = {
            'base_namespace': 'com.jaumo.schema',
            'types_namespace': 'com.jaumo.schema.type',
            'rpc_namespace': 'com.jaumo.schema.rpc',
            'metadata_schema': 'com.jaumo.schema.type.Metadata',
            'schema_mapping_path': ROOT_DIR.joinpath('./schema-mapping.json'),
            'avro_tools_path': AVRO_TOOLS_JAR
        }

        def get_preprocessor(
                input_path: str, input_extension: str, input_format: str,
                output_path: str, output_extension: str, output_format: str) -> AvroPreprocessor:
            paths: Dict[Any, Any] = {**paths_args, **{
                'input_path': str(FIXTURES_DIR.joinpath(input_path)),
                'output_path': str(FIXTURES_DIR.joinpath(output_path)),
                'input_schema_file_extension': input_extension,
                'output_schema_file_extension': output_extension,
                'input_schema_file_format': input_format,
                'output_schema_file_format': output_format,
            }}
            args: Dict[Any, Any] = {**preproc_args, **{'paths': AvroPaths(**paths)}}
            return AvroPreprocessor(**args)

        from avro_preprocessor.modules.documentation_list_condenser import DocumentationCondenser

        # exavsc to yexavsc
        get_preprocessor(
            't3_input/', 'exavsc', 'exavsc',
            't3_yaml/', 'yexavsc', 'yaml'
        ).process([DocumentationCondenser.__name__])

        self.assert_trees_equals('t3_yaml_expected/', 't3_yaml/', 'yexavsc', 'yexavsc')

        # yexavsc to avsc
        get_preprocessor(
            't3_yaml/', 'yexavsc', 'yaml',
            't3_output/', 'avsc', 'json'
        ).process()

        self.assert_trees_equals('t3_expected/', 't3_output/')

        # now mixed sources, exavsc,yexavsc to avsc
        get_preprocessor(
            't3_input_mixed/', 'exavsc,yexavsc', 'yaml',
            't3_output/', 'avsc', 'json'
        ).process()

        self.assert_trees_equals('t3_expected/', 't3_output/')

    def test_full_t4(self) -> None:
        """
        Full test on fixture 't4_input'.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('t4_input/')),
                output_path=str(FIXTURES_DIR.joinpath('t4_output/')),
                base_namespace='com.jaumo.schema',
                types_namespace='com.jaumo.schema.type',
                metadata_schema='com.jaumo.schema.type.Metadata',
                metadata_exclude=['com.jaumo.schema.simple_rpc'],
                input_schema_file_extension='exavsc',
                schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
                schema_mapping_exclude=['com.jaumo.schema.simple_rpc'],
                avro_tools_path=AVRO_TOOLS_JAR
            ),
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process()

        self.assert_trees_equals('t4_expected/', 't4_output/')

    def test_full_t5(self) -> None:
        """
        Full test on fixture 't5_input'.
        Checking that simple fields with logical type user_id are added to the schema mapping
        while collections of user_ids are not.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('t5_input/')),
                output_path=str(FIXTURES_DIR.joinpath('t5_output/')),
                base_namespace='com.jaumo.schema',
                types_namespace='com.jaumo.schema.type',
                metadata_schema='com.jaumo.schema.type.Metadata',
                metadata_exclude=['com.jaumo.schema.simple_rpc'],
                input_schema_file_extension='exavsc',
                schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
                schema_mapping_exclude=['com.jaumo.schema.simple_rpc'],
                schema_mapping_user_id_types_exclude=['array', 'map'],
                avro_tools_path=AVRO_TOOLS_JAR,
            ),
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process()

        with open(ROOT_DIR.joinpath('./schema-mapping.json')) as smf:
            sm = json.load(smf)
        user_id_fields = sm["com.jaumo.schema.domain.user.TestEvent"]["user-id-fields"]
        self.assertEqual(user_id_fields, ['user_id'])

    def test_input_sorting(self) -> None:
        """
        Test on input sorting.
        """
        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths=AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath('t2_input_unsorted/')),
                output_path=str(FIXTURES_DIR.joinpath('t2_sorted/')),
                base_namespace='com.jaumo.event',
                types_namespace='com.jaumo.event.type',
                metadata_schema='com.jaumo.event.type.Metadata',
                input_schema_file_extension='exavsc',
                output_schema_file_extension='exavsc',
                schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
                avro_tools_path=AVRO_TOOLS_JAR
            ),
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process(['AvroSorter'])

        for name, original_schema in avro_preprocessor.schemas.original.items():
            processed_schema = avro_preprocessor.schemas.processed[name]

            original_schema_no_sorting = json.loads(json.dumps(original_schema))
            processed_schema_no_sorting = json.loads(json.dumps(processed_schema))

            self.assertNotEqual(original_schema, processed_schema)
            self.assertEqual(original_schema_no_sorting, processed_schema_no_sorting)

    def assert_trees_equals(self,
                            expected: str,
                            processed: str,
                            expected_extension: str = 'avsc',
                            processed_extension: str = 'avsc',
                            assess_schema_string: bool = True,
                            ) -> None:
        """
        Asserts two schema trees are equal
        :param expected: Path of the expected tree
        :param processed: Path of the processed tree
        :param expected_extension: The extension of files in the expected tree
        :param processed_extension: The extension of files in the processed tree
        :param assess_schema_string: Also assess the string representation of the result
        """
        container_expected = SchemasContainer(
            AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath(expected)),
                output_path="",
                base_namespace='com.jaumo.event',
                input_schema_file_extension=expected_extension
            )
        )
        container_expected.read_schemas()
        expected_schemas = OrderedDict(sorted({
                                                  **container_expected.original,
                                                  **container_expected.autogenerated_keys
                                              }.items()))

        container_processed = SchemasContainer(
            AvroPaths(
                input_path=str(FIXTURES_DIR.joinpath(processed)),
                output_path="",
                base_namespace='com.jaumo.event',
                input_schema_file_extension=processed_extension
            )
        )
        container_processed.read_schemas()
        processed_schemas = OrderedDict(sorted({
                                                   **container_processed.original,
                                                   **container_processed.autogenerated_keys
                                               }.items()))

        self.assertNotEqual(len(expected_schemas), 0)
        self.assertNotEqual(len(processed_schemas), 0)

        expected_schemas_names = set(expected_schemas)
        processed_schemas_names = set(processed_schemas)
        self.assertEqual(expected_schemas_names, processed_schemas_names)

        # self.assertEqual(len(expected_schemas), len(processed_schemas))

        for name, expected_schema in expected_schemas.items():
            print('Asserting', name)
            self.assertEqual(expected_schema, processed_schemas[name])

        if assess_schema_string:
            for name, expected_schema in container_expected.original_string.items():
                print('Asserting string', name)
                self.assertEqual(expected_schema.strip(),  # leading and trailing spaces
                                 container_processed.original_string[name].strip())

    def test_union_field(self) -> None:
        """
        Test the union
        """
        schema = self.build("""
        {
            "name": "int_or_string",
            "type": ["int", "string"],
            "nullable_optional": true,
            "doc": "The address of the user"
        }
        """)

        expected = self.build("""
        {
            "name": "int_or_string",
            "doc": "The address of the user",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalInt_or_string",
                    "doc": "The address of the user (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                "int",
                                "string"
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'OptionalInt_or_string', 'TestRecord'})

    def test_enum_field(self) -> None:
        """
        Test the enum field
        """
        schema = self.build("""
        {
            "name": "kind",
            "type": {
                "name": "Kind",
                "doc": "myenum",
                "type": "enum",
                "symbols": [
                    "ONE",
                    "TWO",
                    "THREE"
                ]
            },
            "nullable_optional": true,
            "doc": "The foo enum"
        }
        """)

        expected = self.build("""
        {
            "name": "kind",
            "doc": "The foo enum",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalKind",
                    "doc": "The foo enum (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "name": "Kind",
                                    "doc": "myenum",
                                    "type": "enum",
                                    "symbols": [
                                        "ONE",
                                        "TWO",
                                        "THREE"
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'Kind', 'OptionalKind', 'TestRecord'})

    def test_union_enum(self) -> None:
        """
        Test enum inside union. The expected result is THE SAME of the previous case.
        """
        schema = self.build("""
        {
            "name": "kind",
            "type": [
                "null",
                {
                    "name": "Kind",
                    "doc": "the kind enum values",
                    "type": "enum",
                    "symbols": [
                        "ONE",
                        "TWO",
                        "THREE"
                    ]
                }
            ],
            "nullable_optional": true,
            "doc": "The foo enum"
        }
        """)

        expected = self.build("""
        {
            "name": "kind",
            "doc": "The foo enum",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalKind",
                    "doc": "The foo enum (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "name": "Kind",
                                    "doc": "the kind enum values",
                                    "type": "enum",
                                    "symbols": [
                                        "ONE",
                                        "TWO",
                                        "THREE"
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'Kind', 'OptionalKind', 'TestRecord'})

    def test_union_record(self) -> None:
        """
        Test records inside union.
        """
        schema = self.build("""
        {
            "name": "request_response",
            "doc": "the request response",
            "nullable_optional": true,
            "type": [
                {
                    "name": "RecordRequest",
                    "doc": "the request",
                    "type": "record",
                    "fields": [
                        {
                            "name": "request_id",
                            "doc": "the request id",
                            "type": "int"
                        },
                        {
                            "name": "message_type",
                            "doc": "the message type",
                            "type": "int"
                        },
                        {
                            "name": "users",
                            "doc": "the users",
                            "type": "string"
                        }
                    ]
                },
                {
                    "name": "RecordResponse",
                    "doc": "the response",
                    "type": "record",
                    "fields": [
                        {
                            "name": "request_id",
                            "doc": "the response id",
                            "type": "int"
                        },
                        {
                            "name": "response_code",
                            "doc": "the code",
                            "type": "string"
                        },
                        {
                            "name": "response_count",
                            "doc": "the count",
                            "type": "int"
                        },
                        {
                            "name": "reason_code",
                            "doc": "the reason",
                            "type": "string"
                        }
                    ]
                }
            ]
        }
        """)

        expected = self.build("""
        {
            "name": "request_response",
            "doc": "the request response",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalRequest_response",
                    "doc": "the request response (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "name": "RecordRequest",
                                    "doc": "the request",
                                    "type": "record",
                                    "fields": [
                                        {
                                            "name": "request_id",
                                            "doc": "the request id",
                                            "type": "int"
                                        },
                                        {
                                            "name": "message_type",
                                            "doc": "the message type",
                                            "type": "int"
                                        },
                                        {
                                            "name": "users",
                                            "doc": "the users",
                                            "type": "string"
                                        }
                                    ]
                                },
                                {
                                    "name": "RecordResponse",
                                    "doc": "the response",
                                    "type": "record",
                                    "fields": [
                                        {
                                            "name": "request_id",
                                            "doc": "the response id",
                                            "type": "int"
                                        },
                                        {
                                            "name": "response_code",
                                            "doc": "the code",
                                            "type": "string"
                                        },
                                        {
                                            "name": "response_count",
                                            "doc": "the count",
                                            "type": "int"
                                        },
                                        {
                                            "name": "reason_code",
                                            "doc": "the reason",
                                            "type": "string"
                                        }
                                    ]
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {
            'OptionalRequest_response',
            'RecordRequest',
            'RecordResponse',
            'TestRecord'
        })

    def test_array(self) -> None:
        """
        Test array.
        """
        schema = self.build("""
        {
            "name":"children",
            "doc":"test array",
            "nullable_optional": true,
            "type":{
                "type": "array",  
                "items":{
                    "name":"Child",
                    "doc":"child",
                    "type":"record",
                    "fields":[
                        {
                            "name":"name",
                            "doc": "my name",
                            "type":"string"
                        }
                    ]
                }
            }
        }
        """)

        expected = self.build("""
        {
            "name": "children",
            "doc": "test array",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalChildren",
                    "doc": "test array (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "type": "array",
                                    "items": {
                                        "name": "Child",
                                        "doc": "child",
                                        "type": "record",
                                        "fields": [
                                            {
                                                "name": "name",
                                                "doc": "my name",
                                                "type": "string"
                                            }
                                        ]
                                    }
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'Child', 'OptionalChildren', 'TestRecord'})

    def test_map(self) -> None:
        """
        Test map.
        """
        schema = self.build("""
        {
            "name":"children",
            "doc":"the map",
            "nullable_optional": true,
            "type": {
                "type": "map", 
                "values": "string"
            }
        }
        """)

        expected = self.build("""
        {
            "name": "children",
            "doc": "the map",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalChildren",
                    "doc": "the map (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "type": "map",
                                    "values": "string"
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'OptionalChildren', 'TestRecord'})

    def test_fixed(self) -> None:
        """
        Test fixed.
        """
        schema = self.build("""
        {
            "name":"children",
            "doc":"the kids",
            "nullable_optional": true,
            "type": {
                "type": "fixed", 
                "doc": "fixed field",
                "size": 16, 
                "name": "md5"
            }
        }
        """)

        expected = self.build("""
        {
            "name": "children",
            "doc": "the kids",
            "default": null,
            "type": [
                "null",
                {
                    "name": "OptionalChildren",
                    "doc": "the kids (Optional Value)",
                    "type": "record",
                    "fields": [
                        {
                            "name": "value",
                            "doc": "The optional value",
                            "type": [
                                "null",
                                {
                                    "name": "md5",
                                    "doc": "fixed field",
                                    "type": "fixed",
                                    "size": 16
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        """)

        self.assert_extension(schema, expected, {'OptionalChildren', 'TestRecord', 'md5'})

    def assert_extension(self, schema: str, expected: str, java_classes_names: Set[str]) -> None:
        """
        Asserts that the extension works.

        :param schema: Schema in input
        :param expected: First (and only) field of the extended schema
        :param java_classes_names: expected names of created java classes
        """
        AvroPaths.reset_directory(str(FIXTURES_DIR.joinpath('/tmp/')))
        filename = 'tmp/input/com/jaumo/event/domain/user/users/TestRecord.exavsc'
        test_record_path = Path(FIXTURES_DIR.joinpath(filename))
        test_record_path.parent.mkdir(parents=True, exist_ok=True)
        test_record_path.write_text(schema)

        paths = AvroPaths(
            input_path=str(FIXTURES_DIR.joinpath('tmp/input/')),
            output_path=str(FIXTURES_DIR.joinpath('tmp/output/')),
            base_namespace='com.jaumo.event',
            types_namespace='com.jaumo.event.type',
            metadata_schema='com.jaumo.event.type.Metadata',
            input_schema_file_extension='exavsc',
            schema_mapping_path=ROOT_DIR.joinpath('./schema-mapping.json'),
            avro_tools_path=AVRO_TOOLS_JAR
        )

        avro_preprocessor: AvroPreprocessor = AvroPreprocessor(
            paths,
            verbose=True,
            json_indent=4,
            schema_registry_url=self.schema_registry_url
        )
        avro_preprocessor.process()

        filename = 'tmp/output/com/jaumo/event/domain/user/users/TestRecord.avsc'
        test_record_path = Path(FIXTURES_DIR.joinpath(filename))
        output_schema = test_record_path.read_text()
        self.assertEqual(output_schema, expected)
        self.assertEqual(
            sorted(java_classes_names),
            JavaClassesCreator(avro_preprocessor.schemas).get_java_classes_names()
        )

    def build(self, fields: str) -> str:
        """
        Create a schema given its fields
        :param fields: the fields
        :return: The schema
        """
        self.delete_test_record_subject()
        schema = self.schema_header + fields + self.schema_footer
        return json.dumps(json.loads(schema, object_pairs_hook=OrderedDict), indent=4)

    def delete_test_record_subject(self) -> None:
        """
        Deletes the subject associated to TestRecord in the Schema Registry because
        it is used multiple times with incompatible schemas
        """
        subject_to_delete = "domain.user.users-com.jaumo.event.domain.user.users.TestRecord"
        url = self.schema_registry_url + "/subjects/" + subject_to_delete
        import requests
        requests.delete(url)



"""
A module to check that every field has appropriate documentation and duplications
due to copy/paste.
"""
import copy
import json
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, Tuple

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.modules.avro_sorter import AvroSorter
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


@dataclass
class Schema:
    """
    Helper class to store schemas
    """
    name: str
    schema: str


class DocumentationChecker(PreprocessorModule):
    """
    Checks documentation fields ('doc').
    """

    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)

        self.failure_due_to_naming = False
        self.current_docs: Dict[str, Schema] = {}

        self.documented_nodes: Tuple = (Avro.Name, Avro.Protocol, Avro.Request, Avro.Response)

    def process(self) -> None:
        """Process all schemas."""

        for _, schema in self.processed_schemas_iter():
            self.traverse_schema(self.check_documentation, schema)

        if self.failure_due_to_naming:
            raise ValueError("Naming conventions are mandatory, see above for specific errors.")

    def check_documentation(self, schema: Avro.Node) -> None:
        """
        Checks that a (sub) schema complies with the documentation policies:

        1. (sub) schemas with a "name" property MUST have a "doc" property.
        2. (sub) schemas without a "name" property MUST NOT have a "doc" property.
        3. If two (sub) schemas have the same "doc", they must be identical.

        :param schema: The (sub) schema
        """

        if isinstance(schema, OrderedDict):

            if Avro.Doc in schema:
                if all(field not in schema for field in self.documented_nodes):
                    self.failure_due_to_naming = True
                    print(
                        'Error: sub-schema with no "name" property has "doc":   ',
                        self.current_schema_name,
                        '   subschema: ', self.json_of(schema)
                    )

                doc = schema[Avro.Doc]

                if doc not in self.current_docs:
                    self.current_docs[doc] = Schema(self.current_schema_name, self.json_of(schema))
                else:
                    subschema_for_this_doc = self.current_docs[doc].schema
                    if subschema_for_this_doc != self.json_of(schema):
                        self.failure_due_to_naming = True
                        print(
                            'Error: duplicated "doc" field for two different (sub)schemas:\n\t',
                            self.current_docs[doc].name, '   ', subschema_for_this_doc, '\n\t',
                            self.current_schema_name, '   ', self.json_of(schema)
                        )

            else:
                if any(field in schema for field in self.documented_nodes):
                    self.failure_due_to_naming = True
                    print(
                        'Error: no "doc" field in schema   ',
                        self.current_schema_name,
                        '   subschema: ', self.json_of(schema)
                    )

    @staticmethod
    def json_of(schema: OrderedDict) -> str:
        """
        Gets a compact, sorted JSON representation of a (sub) schema
        :param schema: The (sub) schema
        :return: The string representation
        """
        schema_copy = copy.deepcopy(schema)
        if Avro.PartitionKey in schema_copy:
            schema_copy.pop(Avro.PartitionKey)
        return json.dumps(AvroSorter.sort_avro(schema_copy), indent=None)



"""
Automatically generates Avro Keys for Kafka topics keys using the property 'partition-field'.
"""
from collections import OrderedDict
from typing import Dict, List, Any

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.avro_naming import AvroNaming
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class KeysGenerator(PreprocessorModule):
    """
    Generates Avro Keys for all topics.
    """

    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)
        self.schema_keys: Dict[str, Any] = {}

        self.admissible_types: List[str] = ['int', 'long', 'string', 'bytes']

    def process(self) -> None:
        """Process all schemas."""

        for name, schema in self.processed_schemas_iter():
            if self.schemas.paths.is_event_resource(name):
                key_name = AvroNaming.get_key_fully_qualified_name(name)

                if key_name not in self.schemas.autogenerated_keys:
                    # one key per topic

                    self.schema_keys[name] = []
                    self.traverse_schema(self.find_partition_keys_records, schema)
                    self.schema_keys[name] = sorted(self.schema_keys[name],
                                                    key=lambda k: k[Avro.Name])  # type: ignore

                    if self.schema_keys[name]:
                        avro_key = OrderedDict((
                            (Avro.Namespace, schema[Avro.Namespace]),
                            (Avro.Name, 'Key'),
                            (Avro.Doc, '{} {}'.format(
                                AvroNaming.autogenerated_key_doc,
                                AvroNaming.get_topic(self.schemas.paths.base_namespace, name))),
                            (Avro.Type, Avro.Record),
                            (Avro.Fields, self.schema_keys[name])
                        ))

                        self.schemas.autogenerated_keys[key_name] = avro_key

    def find_partition_keys_records(
            self, record: Avro.Node) -> None:
        """
        Finds property 'partition-field' inside schemas and saves the record they belong to.
        :param record: The schema
        """
        if isinstance(record, OrderedDict) and Avro.PartitionKey in record:
            if record[Avro.PartitionKey]:

                # type is one of admissible types or a logicalType of admissible types
                if record[Avro.Type] in self.admissible_types or (
                        # type is a logicalType
                        isinstance(record[Avro.Type], OrderedDict)
                        and Avro.Type in record[Avro.Type]
                        and Avro.LogicalType in record[Avro.Type]
                        and record[Avro.Type][Avro.Type] in self.admissible_types
                ):
                    if record not in self.schema_keys[self.current_schema_name]:
                        self.schema_keys[self.current_schema_name].append(record)
                else:
                    raise ValueError(
                        'Partition key in field {} of schema {} of type {}: '
                        'Only "int", "long", "string", or "bytes" are allowed'
                        .format(record[Avro.Name], self.current_schema_name, record[Avro.Type])
                    )



#!/usr/bin/env python3
"""
Script transforming Extended Avro (.exavsc) to regular Avro (.avsc) schema files.
Extends Avro schema definition with 'nullable_optional' keyword.
"""
from collections import OrderedDict
from typing import List, Optional, Union

from avro_preprocessor.avro_paths import AvroPaths
from avro_preprocessor.colored_json import ColoredJson
from avro_preprocessor.modules.avro_sorter import AvroSorter, AvroOrderChecker
from avro_preprocessor.modules.documentation_checker import DocumentationChecker
from avro_preprocessor.modules.documentation_list_condenser import DocumentationCondenser
from avro_preprocessor.modules.java_classes_creator import JavaClassesCreator
from avro_preprocessor.modules.keys_generator import KeysGenerator
from avro_preprocessor.modules.metadata_adder import MetadataAdder
from avro_preprocessor.modules.names_checker import NamesChecker
from avro_preprocessor.modules.namespace_checker import NamespaceChecker
from avro_preprocessor.modules.nullable_optional_expander import NullableOptionalExpander
from avro_preprocessor.modules.references_expander import ReferencesExpander
from avro_preprocessor.modules.schema_dependencies_checker import SchemaDependenciesChecker
from avro_preprocessor.modules.schema_mapping_generator import SchemaMappingGenerator
from avro_preprocessor.modules.schema_registrar import SchemaRegistrar
from avro_preprocessor.modules.schema_registry_checker import SchemaRegistryChecker
from avro_preprocessor.preprocessor_module import PreprocessorModule  # pylint: disable=W0611
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class AvroPreprocessor:
    """
    Avro extender main class.
    """

    #: When adding a module, add it here to make it available
    preprocessing_modules = [
        DocumentationCondenser,
        NamespaceChecker,
        DocumentationChecker,
        NamesChecker,
        AvroOrderChecker,
        MetadataAdder,
        SchemaDependenciesChecker,
        ReferencesExpander,
        NullableOptionalExpander,
        AvroSorter,
        KeysGenerator,
        SchemaMappingGenerator,
        JavaClassesCreator,
        SchemaRegistryChecker,
        SchemaRegistrar
    ]

    #: OrderedDict of module_name -> module_class
    available_preprocessing_modules = OrderedDict((
        (m.__name__, m) for m in preprocessing_modules
    ))

    def __init__(self,
                 paths: AvroPaths,
                 verbose: bool = True,
                 json_indent: int = 4,
                 yaml_indent: Union[dict, None] = None,
                 schema_registry_url: str = 'http://localhost:8081') -> None:
        """
        Initialization function.

        :param paths: The paths of the input and output directories
        :param verbose: Verbose flag
        :param json_indent: How much to indent JSON strings
        :param yaml_indent: How much to indent YAML strings
        :param schema_registry_url: The URL of the Confluent Schema Registry
        """
        if yaml_indent is None:
            yaml_indent = {"mapping": 2, "sequence": 4, "offset": 2}
        self.paths = paths
        self.verbose = verbose
        ColoredJson.json_indent = json_indent
        ColoredJson.yaml_indent = yaml_indent
        SchemaRegistrar.schema_registry_url = schema_registry_url

        self.schemas: SchemasContainer = self._get_schemas_container()

        self.modules: 'OrderedDict[str, PreprocessorModule]' = OrderedDict()

    def _get_schemas_container(self) -> SchemasContainer:
        return SchemasContainer(self.paths, self.verbose)

    def process(self, requested_modules: Optional[List[str]] = None) -> None:
        """
        Process all the schemas using the modules pipeline.
        Standard schemas are saved in .avsc
        (or whatever set by the output_schema_file_extension parameter)
        files in the output path.

        :param requested_modules: The modules to activate, all available ones if set to None
        """

        self.schemas.read_schemas()
        self.modules = self.get_requested_modules(requested_modules)
        self._do_process(self.modules)
        self.schemas.write_schemas()

    def _do_process(self, requested_modules: 'OrderedDict[str, PreprocessorModule]') -> None:
        self.schemas.modules = requested_modules

        for module_name, module in requested_modules.items():
            # self.schemas.print_schemas('json')
            try:
                module.process()
            except Exception as e:
                import sys
                custom_message = f'\n\n{"*" * 100}\nException in module "{module_name}" while ' \
                                 f'processing schema: {module.current_schema_name}\n{"*" * 100}'
                raise type(e)(f'{str(e)} {custom_message}').with_traceback(sys.exc_info()[2])

    def get_requested_modules(self, requested_modules_names: Optional[List[str]]) \
            -> 'OrderedDict[str, PreprocessorModule]':
        """
        Process the requested modules creating an instance of each of them.

        :param requested_modules_names: The names of the requested modules
        :return: List of modules
        """

        if requested_modules_names:
            for module in requested_modules_names:
                if module not in self.available_preprocessing_modules:
                    raise ValueError("No such module {}".format(module))
            requested_modules_names_checked = requested_modules_names
        else:
            requested_modules_names_checked = list(self.available_preprocessing_modules)

        modules: OrderedDict = OrderedDict()
        for module in self.available_preprocessing_modules:
            if module in requested_modules_names_checked:
                mod = self.available_preprocessing_modules[module]
                modules[mod.__name__] = mod(self.schemas)  # type: ignore
        return modules



"""
A module to check that the directory structure corresponds to the schema declared namespace.
"""

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class NamespaceChecker(PreprocessorModule):
    """
    Checks schema namespace.
    """
    def process(self) -> None:
        """Process all schemas."""
        for name, schema in self.processed_schemas_iter():
            class_declared_name = schema[Avro.Namespace] + '.' + \
                                  schema.get(Avro.Name, schema.get(Avro.Protocol))
            # print(name, '=?=', class_declared_name)
            if name != class_declared_name:
                raise ValueError(
                    "File tree path and in schema namespace + name should coincide {} != {}"
                    .format(name, class_declared_name)
                )



from pathlib import Path

from setuptools import setup, find_packages

setup(
    name='avro-preprocessor',
    version_format='{tag}',
    description='A preprocessor for Avro Schemata',
    long_description=Path('README.md').read_text().replace("""![avropreprocessor](avro.jpg)""", ""),
    long_description_content_type='text/markdown',
    keywords=['avro', 'preprocessor', 'schema', 'schemas', 'schemata'],
    url='https://gitlab.com/Jaumo/avro-preprocessor',
    author='Jaumo GmbH',
    author_email='nicola.bova@jaumo.com',
    packages=find_packages(),
    scripts=['avro_preprocessor/avropreprocessor.py'],
    license='Apache2',
    install_requires=[
        'networkx==2.6.2',
        'pygments>=2.6.1',
        'requests==2.20.0',
        'ruamel.yaml==0.16.12',
        'ruamel.yaml.clib==0.2.2',
    ]
)



"""
A module to analyze schema dependencies in a path
"""

from collections import OrderedDict
from typing import Optional

import networkx as nx

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


class SchemaDependenciesChecker(PreprocessorModule):
    """
    Checks schema dependencies - exception thrown if it detects cycles.

    NOTE: Records defined within schemas are ignored.
    """
    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)

        self.record_dependencies: OrderedDict = OrderedDict()
        self.record_dependencies_graph = nx.DiGraph()

        self.current_schema_fully_qualified_name: Optional[str] = None

    def process(self) -> None:
        """Detects all dependencies among schemas."""
        for _, schema in self.processed_schemas_iter():
            self.find_dependencies_in_schema(schema)

        dependencies = {record: sorted(nx.descendants(self.record_dependencies_graph, record))
                        for record in self.record_dependencies_graph}

        self.record_dependencies = \
            OrderedDict(sorted(dependencies.items(), key=lambda kv: len(kv[1])))

        if self.schemas.verbose:
            for idx, (record, dependencies) in enumerate(self.record_dependencies.items()):
                print(idx, '# deps:', len(dependencies), record, dependencies)

        # let's assert there are no cycles
        try:
            cycle = nx.find_cycle(self.record_dependencies_graph, "original")
            print(cycle)
            raise ValueError("The supplied list of schemas contains cyclic dependencies!")
        except nx.exception.NetworkXNoCycle:
            pass  # this is the expected outcome. No cycles should be found.

    def find_dependencies_in_schema(self, schema: OrderedDict) -> None:
        """
        Detects dependencies for a schema.
        :param schema: The schema
        """
        self.current_schema_fully_qualified_name = \
            '.'.join((schema[Avro.Namespace], schema.get(Avro.Name, schema.get(Avro.Protocol))))
        self.record_dependencies_graph.add_node(self.current_schema_fully_qualified_name)

        self.traverse_schema(self.store_dependencies_of_field, schema)

    def store_dependencies_of_field(self, node: Avro.Node) -> None:
        """
        Store dependencies of other records in a node in a private dict
        :param node: The input node
        """

        if isinstance(node, str):
            if '.' in node and node in self.schemas.processed:
                self.record_dependencies_graph.add_edge(
                    self.current_schema_fully_qualified_name, node)



"""
A module to check that every property ('name') has appropriate format.
"""
import json
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict

from avro_preprocessor.avro_domain import Avro
from avro_preprocessor.modules.avro_sorter import AvroSorter
from avro_preprocessor.preprocessor_module import PreprocessorModule
from avro_preprocessor.schemas_container import SchemasContainer

__author__ = "Nicola Bova"
__copyright__ = "Copyright 2018, Jaumo GmbH"
__email__ = "nicola.bova@jaumo.com"


@dataclass
class Schema:
    """
    Helper class to store schemas
    """
    name: str
    schema: str


class NamesChecker(PreprocessorModule):
    """
    Checks property name fields ('name').
    """
    def __init__(self, schemas: SchemasContainer):
        super().__init__(schemas)

        self.current_schema_name: str = ""
        self.failure_due_to_naming = False
        self.current_docs: Dict[str, Schema] = {}

    def process(self) -> None:
        """Process all schemas."""

        for _, schema in self.processed_schemas_iter():
            self.traverse_schema(self.check_property_names, schema)

        if self.failure_due_to_naming:
            raise ValueError("Naming conventions are mandatory, see above for specific errors.")

    def check_property_names(self, schema: Avro.Node) -> None:
        """
        Checks that a (sub) schema complies with the property naming policy.

        :param schema: The (sub) schema
        """

        if not isinstance(schema, OrderedDict):
            return

        if Avro.Protocol in schema:
            if not self.is_camel_case(schema[Avro.Protocol]):
                self.failure_due_to_naming = True
                print(
                    'Error: schema "protocol" does not comply to policy (camel case):   ',
                    self.current_schema_name,
                    '   subschema: ', self.json_of(schema)
                )

        if Avro.Name in schema:
            if Avro.Type in schema and schema[Avro.Type] in (Avro.Record, Avro.Enum, Avro.Error):
                if not self.is_camel_case(schema[Avro.Name]):
                    self.failure_due_to_naming = True
                    print(
                        'Error: schema/record "name" does not comply to policy (camel case):   ',
                        self.current_schema_name,
                        '   subschema: ', self.json_of(schema)
                    )

            elif not self.is_snake_case(schema[Avro.Name]):
                self.failure_due_to_naming = True
                print(
                    'Error: sub-schema "name" property does not comply to policy (snake_case):   ',
                    self.current_schema_name,
                    '   subschema: ', self.json_of(schema)
                )

    @staticmethod
    def is_snake_case(string: str) -> bool:
        """
        Check if a string is snake_case.

        :param string: The string in input
        :return: bool
        """
        return string == string.lower()

    @staticmethod
    def is_camel_case(string: str) -> bool:
        """
        Check if a string is CamelCase.

        :param string: The string in input
        :return: bool
        """
        return '_' not in string

    @staticmethod
    def json_of(schema: OrderedDict) -> str:
        """
        Gets a compact, sorted JSON representation of a (sub) schema
        :param schema: The (sub) schema
        :return: The string representation
        """
        return json.dumps(AvroSorter.sort_avro(schema), indent=None)
